@@ -24,3 +24,10 @@ from __future__ import absolute_import
 from .eam import EAM
 from .pair_potential import PairPotential
 from .supercell_calculator import SupercellCalculator
+from .polydisperse import Polydisperse
+
+try:
+    import scipy.sparse as sp
+except ImportError:
+    warnings.warn('Warning: no scipy')
+
