@@ -4,6 +4,7 @@
 #
 # Copyright (2014) James Kermode, King's College London
 #                  Lars Pastewka, Karlsruhe Institute of Technology
+#                  Wolfram Nöhring, University of Freiburg
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -18,10 +19,7 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 # ======================================================================
-
-"""
-Embedded-atom method potential.
-"""
+"""EAM calculator"""
 
 import os
 
@@ -37,7 +35,13 @@ except:
 
 from scipy.sparse import bsr_matrix
 from matscipy.calculators.eam.io import read_eam
-from matscipy.neighbours import neighbour_list, first_neighbours, find_indices_of_reversed_pairs
+from matscipy.neighbours import (
+    neighbour_list, 
+    first_neighbours, 
+    find_indices_of_reversed_pairs,
+    find_common_neighbours
+)
+
 
 ###
 
@@ -313,7 +317,7 @@ class EAM(Calculator):
         ----------
         atoms : ase.Atoms
         divide_by_masses : bool
-            Divide block :math:`\nu\mu` by :math:`m_\num_\mu` to obtain the dynamical matrix
+            Divide block :math:`\nu\mu` by :math:`m_\nu{}m_\mu{}` to obtain the dynamical matrix
 
         Returns
         -------
@@ -331,7 +335,7 @@ class EAM(Calculator):
          * :math:`r_{\nu\mu{}i} = r_{\mu{}i}-r_{\nu{}i}` 
          * :math:`r_{\nu\mu{}}` Norm of :math:`r_{\nu\mu{}i}`, i.e.\ :math:`\left(r_{\nu\mu{}1}^2+r_{\nu\mu{}2}^2+r_{\nu\mu{}3}^2\right)^{1/2}`
          * :math:`\phi_{\nu\mu}(r_{\nu\mu{}})` Pair potential energy of atoms :math:`\nu` and :math:`\mu` 
-         * :math:`\rho_nu` Total electron density of atom :math:`\nu`  
+         * :math:`\rho_{\nu}` Total electron density of atom :math:`\nu`  
          * :math:`U_\nu(\rho_nu)` Embedding energy of atom :math:`\nu` 
          * :math:`g_{\delta}\left(r_{\gamma\delta}\right) \equiv g_{\gamma\delta}` Contribution from atom :math:`\delta` to :math:`\rho_\gamma`
          * :math:`m_\nu` mass of atom :math:`\nu`
@@ -356,22 +360,8 @@ class EAM(Calculator):
         # with _i indicate arrays containing one element for each atom.
         i_n, j_n, dr_nc, abs_dr_n = neighbour_list('ijDd', atoms,
                                                    self._db_cutoff)
-        first_i = first_neighbours(nat, i_n)
-        # Make sure that the neighborlist does not contain the same pairs twice.
-        # Reoccuring entries may be due to small system size. In this case, the
-        # Hessian matrix will not be symmetric.
-        unique_pairs = set((i, j) for (i, j) in zip(i_n, j_n))
-        if len(unique_pairs) != len(i_n):
-            raise ValueError("neighborlist contains some pairs more than once") 
-        assert(np.all(i_n != j_n))
 
-        if divide_by_masses: 
-            masses_i = atoms.get_masses().reshape(-1, 1, 1)
-            geom_mean_mass_n = np.sqrt(
-                np.take(masses_i, i_n) * np.take(masses_i, j_n)
-            ).reshape(-1, 1, 1)
-
-        # Calculate the derivatives of the pair energy
+        # Calculate derivatives of the pair energy
         drep_n = np.zeros_like(abs_dr_n)  # first derivative
         ddrep_n = np.zeros_like(abs_dr_n) # second derivative
         for atidx1, atnum1 in enumerate(self._db_atomic_numbers):
@@ -389,9 +379,7 @@ class EAM(Calculator):
                         r = rep(abs_dr_n[mask])/abs_dr_n[mask]
                         drep_n[mask] = (drep(abs_dr_n[mask])-r) / abs_dr_n[mask]
                         ddrep_n[mask] = (ddrep(abs_dr_n[mask]) - 2.0 * drep_n[mask]) / abs_dr_n[mask]
-
-        # Calculate the total electron density at each atom, and the 
-        # derivatives of pairwise contributions
+        # Calculate electron density and its derivatives
         f_n = np.zeros_like(abs_dr_n)
         df_n = np.zeros_like(abs_dr_n)  # first derivative
         ddf_n = np.zeros_like(abs_dr_n) # second derivative
@@ -417,7 +405,6 @@ class EAM(Calculator):
                     ddf_n[mask1] = ddf1(abs_dr_n[mask1])
         # Accumulate density contributions
         density_i = np.bincount(i_n, weights=f_n, minlength=nat)
-
         # Calculate the derivatives of the embedding energy
         demb_i = np.zeros(nat)   # first derivative
         ddemb_i = np.zeros(nat)  # second derivative
@@ -430,15 +417,131 @@ class EAM(Calculator):
                 demb_i[mask] += dF(density_i[mask])
                 ddemb_i[mask] += ddF(density_i[mask])
 
-        symmetry_check = False # check symmetry of individual terms?
+        # There are two ways to divide the Hessian by atomic masses, either
+        # during or after construction. The former is preferable with regard
+        # to memory consumption. If we would divide by masses afterwards,
+        # we would have to create a sparse matrix with the same size as the
+        # Hessian matrix, i.e. we would momentarily need twice the given memory.
+        if divide_by_masses: 
+            masses_i = atoms.get_masses().reshape(-1, 1, 1)
+            geom_mean_mass_n = np.sqrt(
+                np.take(masses_i, i_n) * np.take(masses_i, j_n)
+            ).reshape(-1, 1, 1)
+        else:
+            masses_i = None
+            geom_mean_mass_n = None
+
         #------------------------------------------------------------------------ 
         # Calculate pair contribution to the Hessian matrix
         #------------------------------------------------------------------------ 
+        first_i = first_neighbours(nat, i_n)
         e_nc = (dr_nc.T / abs_dr_n).T # normalized distance vectors r_i^{\mu\nu}
-        outer_1 = e_nc.reshape(-1, 3, 1) * e_nc.reshape(-1, 1, 3)
-        outer_2 = np.eye(3, dtype=e_nc.dtype) - outer_1
-        D_ncc = -(ddrep_n * outer_1.T).T
-        D_ncc += -(drep_n / abs_dr_n * outer_2.T).T
+        outer_e_ncc = e_nc.reshape(-1, 3, 1) * e_nc.reshape(-1, 1, 3)
+        eye_minus_outer_e_ncc = np.eye(3, dtype=e_nc.dtype) - outer_e_ncc
+        D = self._calculate_hessian_pair_term(
+            nat, i_n, j_n, abs_dr_n, first_i, 
+            drep_n, ddrep_n, outer_e_ncc, eye_minus_outer_e_ncc, 
+            divide_by_masses, geom_mean_mass_n, masses_i
+        )
+        drep_n = None
+        ddrep_n = None
+
+        #------------------------------------------------------------------------ 
+        # Calculate contribution of embedding term
+        #------------------------------------------------------------------------ 
+        # For each pair in the neighborlist, create arrays which store the 
+        # derivatives of the embedding energy of the corresponding atoms.
+        demb_i_n = np.take(demb_i, i_n)
+        demb_j_n = np.take(demb_i, j_n)
+
+        # Let r be an index into the neighbor list. df_n[r] contains the the
+        # contribution from atom j_n[r] to the derivative of the electron
+        # density of atom i_n[r]. We additionally need the contribution of
+        # i_n[r] to the derivative of j_n[r]. This value is also in df_n,
+        # but at a different position. reverse[r] gives the new index s
+        # where we find this value. The same indexing applies to ddf_n.
+        reverse = find_indices_of_reversed_pairs(i_n, j_n, abs_dr_n)
+        df_i_n = np.take(df_n, reverse)
+        ddf_i_n = np.take(ddf_n, reverse)
+        #we already have ddf_j_n = ddf_n 
+        reverse = None
+
+        df_n_e_nc_outer_product = (df_n * e_nc.T).T
+        df_e_ni = np.empty((nat, 3), dtype=df_n.dtype)
+        for x in range(3):
+            df_e_ni[:, x] = np.bincount(
+                i_n, weights=df_n_e_nc_outer_product[:, x], minlength=nat
+            )
+        df_n_e_nc_outer_product = None
+
+        D += self._calculate_hessian_embedding_term_1(
+            nat, ddemb_i, df_e_ni, 
+            divide_by_masses, masses_i
+        )
+        D += self._calculate_hessian_embedding_term_2(
+            nat, j_n, first_i, ddemb_i, df_i_n, e_nc, df_e_ni, 
+            divide_by_masses, geom_mean_mass_n
+        )
+        D += self._calculate_hessian_embedding_term_3(
+            nat, i_n, j_n, first_i, ddemb_i, df_n, e_nc, df_e_ni, 
+            divide_by_masses, geom_mean_mass_n
+        )
+        df_e_ni = None
+        D += self._calculate_hessian_embedding_terms_4_and_5(
+            nat, first_i, i_n, j_n, outer_e_ncc, demb_i_n, demb_j_n, ddf_i_n, ddf_n, 
+            divide_by_masses, masses_i, geom_mean_mass_n
+        )
+        outer_e_ncc = None
+        ddf_i_n = None
+        ddf_n = None
+        D += self._calculate_hessian_embedding_terms_6_and_7(
+            nat, i_n, j_n, first_i, abs_dr_n, eye_minus_outer_e_ncc, demb_i_n, demb_j_n, df_n, df_i_n,
+            divide_by_masses, masses_i, geom_mean_mass_n
+        )
+        eye_minus_outer_e_ncc = None
+        df_n = None
+        demb_i_n = None
+        demb_j_n = None
+        abs_dr_n = None
+        D += self._calculate_hessian_embedding_term_8(
+            nat, i_n, j_n, e_nc, ddemb_i, df_i_n,
+            divide_by_masses, masses_i, geom_mean_mass_n
+        )
+        return D
+
+    def _calculate_hessian_pair_term(
+        self, nat, i_n, j_n, abs_dr_n, first_i, drep_n, ddrep_n, outer_e_ncc, eye_minus_outer_e_ncc, 
+        divide_by_masses=False, geom_mean_mass_n=None, masses_i=None):
+        """Calculate the pair energy contribution to the Hessian.
+
+        Parameters
+        ----------
+        nat : int
+            number of atoms
+        i_n, j_n : array_like
+            Neighbor pairs
+        abs_dr_n : array_like
+            Length of distance vectors between neighbors
+        first_i : array_like
+            Indices in :code:`i_n` where contiguous 
+            blocks with the same value start
+        drep_n : array_like
+            First derivative of the pair energy with
+            respect to distance vectors between neighbors
+        ddrep_n : array_like
+            Second derivative of the pair energy with
+            respect to distance vectors between neighbors
+        outer_e_ncc : array_like
+            outer product of normalized distance vectors
+        eye_minus_outer_e_ncc : array_like
+            identity matrix minus outer product of normalized distance vectors
+
+        Returns
+        -------
+        D : scipy.sparse.bsr_matrix
+        """
+        D_ncc = -(ddrep_n * outer_e_ncc.T).T
+        D_ncc += -(drep_n / abs_dr_n * eye_minus_outer_e_ncc.T).T
         if divide_by_masses:
             D = bsr_matrix(
                 (D_ncc / geom_mean_mass_n, j_n, first_i), 
@@ -454,89 +557,219 @@ class EAM(Calculator):
             Ddiag /= masses_i
         # put 3x3 blocks on diagonal (Kronecker Delta delta_{\mu\nu})
         D += bsr_matrix((Ddiag, np.arange(nat), np.arange(nat+1)), shape=(3*nat, 3*nat))
+        return D
+    
+    def _calculate_hessian_embedding_term_1(self, nat, ddemb_i, df_e_ni, 
+        divide_by_masses=False, masses_i=None, symmetry_check=False):
+        """Calculate term 1 in the embedding part of the Hessian matrix.
 
-        #------------------------------------------------------------------------ 
-        # Calculate contribution of embedding term
-        #------------------------------------------------------------------------ 
-        # For each pair in the neighborlist, create arrays which store the 
-        # derivatives of the embedding energy of the corresponding atoms.
-        demb_i_n = np.take(demb_i, i_n)
-        demb_j_n = np.take(demb_i, j_n)
-        ddemb_i_n = np.take(ddemb_i, i_n)
-        ddemb_j_n = np.take(ddemb_i, j_n)
+        .. math::
 
-        # Let r be an index into the neighbor list. df_n[r] contains the the
-        # contribution from atom j_n[r] to the derivative of the electron
-        # density of atom i_n[r]. We additionally need the contribution of
-        # i_n[r] to the derivative of j_n[r]. This value is also in df_n,
-        # but at a different position. reverse[r] gives the new index s
-        # where we find this value. The same indexing applies to ddf_n.
-        reverse = find_indices_of_reversed_pairs(i_n, j_n, abs_dr_n)
-        df_i_n = np.take(df_n, reverse)
-        ddf_i_n = np.take(ddf_n, reverse)
-        #we already have ddf_j_n = ddf_n 
+            T_{\nu\mu}^{(1)} = \delta_{\nu\mu}U_\nu''
+            \sum_{\gamma\neq\nu}^{\natoms}g_{\nu\gamma}'\frac{r_{\nu\gamma i}}{r_{\nu\gamma}} 
+            \sum_{\gamma\neq\nu}^{\natoms}g_{\nu\gamma}'\frac{r_{\nu\gamma j}}{r_{\nu\gamma}} 
 
-        # Term 1: 
-        # \delta_{\nu\mu}U_\nu''
-        # \sum_{\gamma\neq\nu}^{\natoms}g_{\nu\gamma}'\frac{r_{\nu\gamma i}}{r_{\nu\gamma}} 
-        # \sum_{\gamma\neq\nu}^{\natoms}g_{\nu\gamma}'\frac{r_{\nu\gamma j}}{r_{\nu\gamma}} 
-        # Likely zero in equilibrium because the sum is zero (appears in the force vector)
-        df_n_e_nc_outer_product = (df_n * e_nc.T).T
-        df_n_e_nc_i = np.empty((nat, 3), dtype=df_n.dtype)
-        for x in range(3):
-            df_n_e_nc_i[:, x] = np.bincount(i_n, weights=df_n_e_nc_outer_product[:, x], minlength=nat)
-        term_1_ncc = ((ddemb_i * df_n_e_nc_i.T).T).reshape(-1,3,1) * df_n_e_nc_i.reshape(-1,1,3)
+        This term is likely zero in equilibrium because
+        the sum is zero (appears in the force vector).
+            
+        Parameters
+        ----------
+        nat : int
+            Number of atoms
+        ddemb_i : array_like
+            Second derivative of the embedding energy
+        df_e_ni : array_like
+            First derivative of the electron density times
+            the normalized distance vector between neighbors
+        divide_by_masses : bool
+            Divide term by geometric mean of mass of pairs of atoms
+            to obtain the contribution to the dynamical matrix
+        masses_i : array_like
+            masses of atoms :code:`i`
+        symmetry_check : bool
+            Check if the term is symmetric
+
+        Returns
+        -------
+        D : scipy.sparse.bsr_matrix
+        """
+        term_1_ncc = ((ddemb_i * df_e_ni.T).T).reshape(-1,3,1) * df_e_ni.reshape(-1,1,3)
         if divide_by_masses:
             term_1_ncc /= masses_i
         term_1 = bsr_matrix((term_1_ncc, np.arange(nat), np.arange(nat+1)), shape=(3*nat, 3*nat)) 
-        D += term_1
         if symmetry_check: 
             print("check term 1", np.linalg.norm(term_1.todense() - term_1.todense().T))
+        return term_1
+    
+    def _calculate_hessian_embedding_term_2(self, nat, j_n, first_i, ddemb_i, df_i_n, e_nc, df_e_ni, 
+        divide_by_masses=False, geom_mean_mass_n=None, symmetry_check=False):
+        """Calculate term 2 in the embedding part of the Hessian matrix.
 
-        # Term 2: 
-        # -u_\nu''g_{\nu\mu}' \frac{r_{\nu\mu j}}{r_{\nu\mu}} \sum_{\gamma\neq\nu}^{\natoms} 
-        #      g_{\nu\gamma}' \frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
-        # Likely zero in equilibrium because the sum is zero (appears in the force vector)
-        df_n_e_nc_j_n = np.take(df_n_e_nc_i, j_n, axis=0)
+        .. math::
+
+            T_{\nu\mu}^{(2)} = 
+            -u_\nu''g_{\nu\mu}' \frac{r_{\nu\mu j}}{r_{\nu\mu}} \sum_{\gamma\neq\nu}^{\natoms} 
+            g_{\nu\gamma}' \frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
+
+        This term is likely zero in equilibrium because
+        the sum is zero (appears in the force vector).
+            
+        Parameters
+        ----------
+        nat : int
+            Number of atoms
+        j_n : array_like
+            Indices of neighbor atoms
+        first_i : array_like
+            Indices in :code:`i_n` where contiguous 
+            blocks with the same value start
+        ddemb_i : array_like
+            Second derivative of the embedding energy
+        df_i_n : array_like
+            Derivative of the electron density of atom :code:`j`
+            with respect to the distance to atom :code:`i`
+        e_nc : array_like
+            Normalized distance vectors between neighbors
+        df_e_ni : array_like
+            First derivative of the electron density times
+            the normalized distance vector between neighbors
+        divide_by_masses : bool
+            Divide term by geometric mean of mass of pairs of atoms
+            to obtain the contribution to the dynamical matrix
+        masses_i : array_like
+            masses of atoms :code:`i`
+        geom_mean_mass_n : array_like
+            geometric mean of masses of pairs of atoms
+        symmetry_check : bool
+            Check if the term is symmetric
+
+        Returns
+        -------
+        D : scipy.sparse.bsr_matrix
+        """
+        df_n_e_nc_j_n = np.take(df_e_ni, j_n, axis=0)
+        ddemb_j_n = np.take(ddemb_i, j_n)
         term_2_ncc = ((ddemb_j_n * df_i_n * e_nc.T).T).reshape(-1,3,1) * df_n_e_nc_j_n.reshape(-1,1,3)
         if divide_by_masses:
             term_2_ncc /= geom_mean_mass_n
         term_2 = bsr_matrix((term_2_ncc, j_n, first_i), shape=(3*nat, 3*nat))
-        D += term_2 
         if symmetry_check: 
             print("check term 2", np.linalg.norm(term_2.todense() - term_2.todense().T))
+        return term_2
+    
+    def _calculate_hessian_embedding_term_3(self, nat, i_n, j_n, first_i, ddemb_i, df_n, e_nc, df_e_ni, 
+        divide_by_masses=False, geom_mean_mass_n=None, symmetry_check=False):
+        """Calculate term 3 in the embedding part of the Hessian matrix.
 
-        # Term 3: 
-        # +u_\mu''g_{\mu\nu}' \frac{r_{\nu\mu i}}{r_{\nu\mu}} \sum_{\gamma\neq\mu}^{\natoms} 
-        #      g_{\mu\gamma}' \frac{r_{\mu\gamma j}}{r_{\mu\gamma}} 
+        .. math::
+
+            T_{\nu\mu}^{(3)} = 
+            +u_\mu''g_{\mu\nu}' \frac{r_{\nu\mu i}}{r_{\nu\mu}} \sum_{\gamma\neq\mu}^{\natoms} 
+            g_{\mu\gamma}' \frac{r_{\mu\gamma j}}{r_{\mu\gamma}} 
+
+        This term is likely zero in equilibrium because
+        the sum is zero (appears in the force vector).
+            
+        Parameters
+        ----------
+        nat : int
+            Number of atoms
+        i_n, j_n : array_like
+            Neighbor pairs
+        first_i : array_like
+            Indices in :code:`i_n` where contiguous 
+            blocks with the same value start
+        ddemb_i : array_like
+            Second derivative of the embedding energy
+        df_n : array_like
+            Derivative of the electron density of atom :code:`i`
+            with respect to the distance to atom :code:`j`
+        e_nc : array_like
+            Normalized distance vectors between neighbors
+        df_e_ni : array_like
+            First derivative of the electron density times
+            the normalized distance vector between neighbors
+        divide_by_masses : bool
+            Divide term by geometric mean of mass of pairs of atoms
+            to obtain the contribution to the dynamical matrix
+        masses_i : array_like
+            masses of atoms :code:`i`
+        geom_mean_mass_n : array_like
+            geometric mean of masses of pairs of atoms
+        symmetry_check : bool
+            Check if the term is symmetric
+
+        Returns
+        -------
+        D : scipy.sparse.bsr_matrix
+        """
         # Likely zero in equilibrium because the sum is zero (appears in the force vector)
-        df_n_e_nc_i_n = np.take(df_n_e_nc_i, i_n, axis=0)
-        term_3_ncc = -((ddemb_i_n * df_n * df_n_e_nc_i_n.T).T).reshape(-1,3,1) * e_nc.reshape(-1,1,3)
+        df_e_ni_n = np.take(df_e_ni, i_n, axis=0)
+        ddemb_i_n = np.take(ddemb_i, i_n)
+        term_3_ncc = -((ddemb_i_n * df_n * df_e_ni_n.T).T).reshape(-1,3,1) * e_nc.reshape(-1,1,3)
         if divide_by_masses:
             term_3_ncc /= geom_mean_mass_n
         term_3 = bsr_matrix((term_3_ncc, j_n, first_i), shape=(3*nat, 3*nat))
-        D += term_3
         if symmetry_check: 
             print("check term 3", np.linalg.norm(term_3.todense() - term_3.todense().T))
+        return term_3
+    
+    def _calculate_hessian_embedding_terms_4_and_5(
+        self, nat, first_i, i_n, j_n, outer_e_ncc, demb_i_n, demb_j_n, ddf_i_n, ddf_n, 
+        divide_by_masses=False, masses_i=None, geom_mean_mass_n=None, symmetry_check=False):
+        """Calculate term 4 and 5 in the embedding part of the Hessian matrix.
 
-        # Term 4:
-        # -\left(u_\mu'g_{\mu\nu}'' + u_\nu'g_{\nu\mu}''\right)
-        # \left(
-        # \frac{r_{\nu\mu i}}{r_{\nu\mu}} 
-        # \frac{r_{\nu\mu j}}{r_{\nu\mu}}
-        # \right)
-        tmp_1 = -((demb_j_n * ddf_i_n + demb_i_n * ddf_n) * outer_1.T).T 
+        .. math::
+
+            T_{\nu\mu}^{(4)} &= -\left(u_\mu'g_{\mu\nu}'' + u_\nu'g_{\nu\mu}''\right)
+            \left(
+            \frac{r_{\nu\mu i}}{r_{\nu\mu}} 
+            \frac{r_{\nu\mu j}}{r_{\nu\mu}}
+            \right)\\
+            T_{\nu\mu}^{(5)} &= \delta_{\nu\mu} \sum_{\gamma\neq\nu}^{N}
+            \left(U_\gamma'g_{\gamma\nu}'' + U_\nu'g_{\nu\gamma}''\right)
+            \left(
+            \frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
+            \frac{r_{\nu\gamma j}}{r_{\nu\gamma}}
+            \right) 
+
+        Parameters
+        ----------
+        nat : int
+            Number of atoms
+        i_n, j_n : array_like
+            Neighbor pairs
+        first_i : array_like
+            Indices in :code:`i_n` where contiguous 
+            blocks with the same value start
+        ddemb_i : array_like
+            Second derivative of the embedding energy
+        df_n : array_like
+            Derivative of the electron density of atom :code:`i`
+            with respect to the distance to atom :code:`j`
+        e_nc : array_like
+            Normalized distance vectors between neighbors
+        df_e_ni : array_like
+            First derivative of the electron density times
+            the normalized distance vector between neighbors
+        divide_by_masses : bool
+            Divide term by geometric mean of mass of pairs of atoms
+            to obtain the contribution to the dynamical matrix
+        masses_i : array_like
+            masses of atoms :code:`i`
+        geom_mean_mass_n : array_like
+            geometric mean of masses of pairs of atoms
+        symmetry_check : bool
+            Check if the terms are symmetric
+
+        Returns
+        -------
+        D : scipy.sparse.bsr_matrix
+        """
+        tmp_1 = -((demb_j_n * ddf_i_n + demb_i_n * ddf_n) * outer_e_ncc.T).T 
         # We don't immediately add term 4 to the matrix, because it would have 
         # to be normalized by the masses if divide_by_masses is true. However,
         # for construction of term 5, we need term 4 without normalization
-
-        # Term 5:
-        # \delta_{\nu\mu} \sum_{\gamma\neq\nu}^{\natoms}
-        #\left(U_\gamma'g_{\gamma\nu}'' + U_\nu'g_{\nu\gamma}''\right)
-        #\left(
-        #\frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
-        #\frac{r_{\nu\gamma j}}{r_{\nu\gamma}}
-        #\right)
         tmp_1_summed = np.empty((nat, 3, 3), dtype=tmp_1.dtype)
         for x in range(3):
             for y in range(3):
@@ -544,34 +777,77 @@ class EAM(Calculator):
         if divide_by_masses:
             tmp_1_summed /= masses_i
         term_5 = bsr_matrix((tmp_1_summed, np.arange(nat), np.arange(nat+1)), shape=(3*nat, 3*nat))
-        D += term_5
         if symmetry_check: 
             print("check term 5", np.linalg.norm(term_5.todense() - term_5.todense().T))
         if divide_by_masses:
             tmp_1 /= geom_mean_mass_n
         term_4 = bsr_matrix((tmp_1, j_n, first_i), shape=(3*nat, 3*nat))
-        D += term_4 
         if symmetry_check: 
             print("check term 4", np.linalg.norm(term_4.todense() - term_4.todense().T))
+        return term_4 + term_5
+    
+    def _calculate_hessian_embedding_terms_6_and_7(
+        self, nat, i_n, j_n, first_i, abs_dr_n, eye_minus_outer_e_ncc, demb_i_n, demb_j_n, df_n, df_i_n,
+        divide_by_masses=False, masses_i=None, geom_mean_mass_n=None, symmetry_check=False):
+        """Calculate term 6 and 7 in the embedding part of the Hessian matrix.
 
-        # Term 6:
-        # -\left(U_\mu'g_{\mu\nu}' + U_\nu'g_{\nu\mu}'\right) \frac{1}{r_{\nu\mu}}
-        #\left(
-        #\delta_{ij}- 
-        #\frac{r_{\nu\mu i}}{r_{\nu\mu}} 
-        #\frac{r_{\nu\mu j}}{r_{\nu\mu}}
-        #\right)
+        .. math::
+
+            T_{\nu\mu}^{(6)} &= -\left(U_\mu'g_{\mu\nu}' + U_\nu'g_{\nu\mu}'\right) \frac{1}{r_{\nu\mu}}
+            \left(
+            \delta_{ij}- 
+            \frac{r_{\nu\mu i}}{r_{\nu\mu}} 
+            \frac{r_{\nu\mu j}}{r_{\nu\mu}}
+            \right) \\
+            T_{\nu\mu}^{(7)}&= \delta_{\nu\mu} \sum_{\gamma\neq\nu}^{N}
+            \left(U_\gamma'g_{\gamma\nu}' + U_\nu'g_{\nu\gamma}'\right) \frac{1}{r_{\nu\gamma}}
+            \left(\delta_{ij}-
+            \frac{r_{\nu\gamma i}}{r_{\nu\gamma}} 
+            \frac{r_{\nu\gamma j}}{r_{\nu\gamma}}
+            \right) 
+
+        Parameters
+        ----------
+        nat : int
+            Number of atoms
+        i_n, j_n : array_like
+            Neighbor pairs
+        first_i : array_like
+            Indices in :code:`i_n` where contiguous 
+            blocks with the same value start
+        abs_dr_n : array_like
+            Length of distance vectors between neighbors
+        eye_minus_outer_e_ncc : array_like
+            identity matrix minus outer product of normalized distance vectors
+        demb_i_n : array_like
+            First derivative of the embedding energy for
+            atoms in the neighbor list
+        demb_j_n : array_like
+            First derivative of the embedding energy of
+            neighbor atoms in the neighbor list
+        df_n : array_like
+            Derivative of the electron density of atom :code:`i`
+            with respect to the distance to atom :code:`j`
+        df_i_n : array_like
+            Derivative of the electron density of atom :code:`j`
+            with respect to the distance to atom :code:`i`
+        divide_by_masses : bool
+            Divide term by geometric mean of mass of pairs of atoms
+            to obtain the contribution to the dynamical matrix
+        masses_i : array_like
+            masses of atoms :code:`i`
+        geom_mean_mass_n : array_like
+            geometric mean of masses of pairs of atoms
+        symmetry_check : bool
+            Check if the terms are symmetric
+
+        Returns
+        -------
+        D : scipy.sparse.bsr_matrix
+        """
         # Like term 4, which was needed to construct term 5, we don't add 
         # term 6 immediately, because it is needed for construction of term 7
-        tmp_2 = -((demb_j_n * df_i_n + demb_i_n * df_n) / abs_dr_n * outer_2.T).T
-
-        # Term 7:
-        # \delta_{\nu\mu} \sum_{\gamma\neq\nu}^{\natoms}
-        #\left(U_\gamma'g_{\gamma\nu}' + U_\nu'g_{\nu\gamma}'\right) \frac{1}{r_{\nu\gamma}}
-        #\left(\delta_{ij}-
-        #\frac{r_{\nu\gamma i}}{r_{\nu\gamma}}
-        #\frac{r_{\nu\gamma j}}{r_{\nu\gamma}}
-        #\right)
+        tmp_2 = -((demb_j_n * df_i_n + demb_i_n * df_n) / abs_dr_n * eye_minus_outer_e_ncc.T).T
         tmp_2_summed = np.empty((nat, 3, 3), dtype=tmp_2.dtype)
         for x in range(3):
             for y in range(3):
@@ -579,74 +855,79 @@ class EAM(Calculator):
         if divide_by_masses:
             tmp_2_summed /= masses_i
         term_7 = bsr_matrix((tmp_2_summed, np.arange(nat), np.arange(nat+1)), shape=(3*nat, 3*nat))
-        D += term_7 
         if symmetry_check: 
             print("check term 7", np.linalg.norm(term_7.todense() - term_7.todense().T))
         if divide_by_masses:
             tmp_2 /= geom_mean_mass_n
         term_6 = bsr_matrix((tmp_2, j_n, first_i), shape=(3*nat, 3*nat))
-        D += term_6 
         if symmetry_check: 
             print("check term 6", np.linalg.norm(term_6.todense() - term_6.todense().T))
+        return term_6 + term_7
+    
+    def _calculate_hessian_embedding_term_8(self, nat, i_n, j_n, e_nc, ddemb_i, df_i_n,
+        divide_by_masses=False, masses_i=None, geom_mean_mass_n=None, symmetry_check=False):
+        """Calculate term 8 in the embedding part of the Hessian matrix.
 
-        #Term 8: 
-        #  \sum_{\substack{\gamma\neq\nu \\ \gamma \neq \mu}}^{\natoms}
-        #U_\gamma'' g_{\gamma\nu}'g_{\gamma\mu}' 
-        #\frac{r_{\gamma\nu i}}{r_{\gamma\nu}}
-        #\frac{r_{\gamma\mu j}}{r_{\gamma\mu}} 
-        #-----------------------------------------------------------------------
-        # This term requires knowledge of common neighbors of pairs of atoms.
-        # Construction of a common neighbor list in Python is likely a
-        # performance bottleneck; it should be implemented in C++ instead.
-        #-----------------------------------------------------------------------
+        .. math::
 
-        # For each (i1, j2)-pair in the neighbor list, find all other
-        # pairs (i2, j) which share the same j. This includes (i1, j)
-        # itself. In this way, create a list with n blocks of rows, where
-        # n is the length of the neighbor list. All rows in a block have
-        # the same j. Each row corresponds to one triplet i1, j, i2.
-        # The number of rows in the block is equal to the total number
-        # of neighbors of j. This is the common neighbor list (cnl).
-        # Besides the block number and i1, j, i2, the cnl stores the
-        # index into the neighbor list where the pair i2-j can be found.
-        j_order = np.argsort(j_n)
-        i_n_2 = i_n[j_order]
-        j_n_2 = j_n[j_order]
-        first_j = first_neighbours(nat, j_n_2)
-        num_rows_per_j = first_j[j_n+1] - first_j[j_n]
-        total_num_rows = np.sum(num_rows_per_j)
-        # The common neighbor information could be stored as
-        # a 2D array. However, multiple 1D arrays are likely
-        # better for performance (fewer cache misses later).
-        cnl_block_number = np.empty(total_num_rows, dtype=i_n.dtype)
-        cnl_j1 = np.empty(total_num_rows, dtype=i_n.dtype)
-        cnl_j_order = np.empty(total_num_rows, dtype=i_n.dtype)
-        cnl_i1_i2 = np.empty((total_num_rows, 2), dtype=i_n.dtype)
-        block_start = np.r_[0, np.cumsum(num_rows_per_j)]
-        slice_for_j1 = {j1: slice(first_j[j1], first_j[j1+1]) for j1 in np.arange(nat)}
-        for block_number, (i1, j1) in enumerate(zip(i_n, j_n)):
-            slice1 = slice(block_start[block_number], block_start[block_number+1])
-            slice2 = slice_for_j1[j1] 
-            cnl_block_number[slice1] = block_number
-            cnl_j1[slice1] = j1 
-            cnl_j_order[slice1] = j_order[slice2]
-            cnl_i1_i2[slice1, 0] = i1
-            cnl_i1_i2[slice1, 1] = i_n_2[slice2]
-        # Determine bins for accumulation by bincount
+            T_{\nu\mu}^{(8)} = 
+            \sum_{\substack{\gamma\neq\nu \\ \gamma \neq \mu}}^{N}
+            U_\gamma'' g_{\gamma\nu}'g_{\gamma\mu}' 
+            \frac{r_{\gamma\nu i}}{r_{\gamma\nu}}
+            \frac{r_{\gamma\mu j}}{r_{\gamma\mu}} 
+
+        This term requires knowledge of common neighbors of pairs of atoms.
+
+        Parameters
+        ----------
+        nat : int
+            Number of atoms
+        i_n, j_n : array_like
+            Neighbor pairs
+        first_i : array_like
+            Indices in :code:`i_n` where contiguous 
+            blocks with the same value start
+        ddemb_i : array_like
+            Second derivative of the embedding energy
+        e_nc : array_like
+            Normalized distance vectors between neighbors
+        df_i_n : array_like
+            Derivative of the electron density of atom :code:`j`
+            with respect to the distance to atom :code:`i`
+        divide_by_masses : bool
+            Divide term by geometric mean of mass of pairs of atoms
+            to obtain the contribution to the dynamical matrix
+        masses_i : array_like
+            masses of atoms :code:`i`
+        geom_mean_mass_n : array_like
+            geometric mean of masses of pairs of atoms
+        symmetry_check : bool
+            Check if the terms are symmetric
+
+        Returns
+        -------
+        D : scipy.sparse.bsr_matrix
+        """
+        cnl_i1_i2, cnl_j1, nl_index_i1_j1, nl_index_i2_j1 = find_common_neighbours(i_n, j_n, nat)
         unique_pairs_i1_i2, bincount_bins = np.unique(cnl_i1_i2, axis=0, return_inverse=True)
-        e_nu_cnl = np.take(e_nc, cnl_block_number, axis=0)
-        e_mu_cnl = np.take(e_nc, cnl_j_order, axis=0)
-        ddemb_cnl = np.take(ddemb_i, cnl_j1)
-        df_nu_cnl = np.take(df_i_n, cnl_block_number)
-        df_mu_cnl = np.take(df_i_n, cnl_j_order)
-        tmp_3 = ((ddemb_cnl * df_nu_cnl * df_mu_cnl * e_nu_cnl.T).T).reshape(-1, 3, 1) * e_mu_cnl.reshape(-1, 1, 3)
+        cnl_i1_i2 = None
+        tmp_3 = np.take(df_i_n, nl_index_i1_j1) * np.take(ddemb_i, cnl_j1) * np.take(df_i_n, nl_index_i2_j1)
+        cnl_j1 = None
         tmp_3_summed = np.empty((unique_pairs_i1_i2.shape[0], 3, 3), dtype=e_nc.dtype)
-        for x in range(3):
-            for y in range(3):
-                tmp_3_summed[:, x, y] = np.bincount(
-                        bincount_bins, weights=tmp_3[:, x, y], 
-                        minlength=unique_pairs_i1_i2.shape[0]
-                ) 
+        for x, y in np.ndindex(3, 3):
+            weights = (tmp_3 * 
+                np.take(e_nc[:, x], nl_index_i1_j1) * 
+                np.take(e_nc[:, y], nl_index_i2_j1)
+            )
+            tmp_3_summed[:, x, y] = np.bincount(
+                    bincount_bins, weights=weights, 
+                    minlength=unique_pairs_i1_i2.shape[0]
+            ) 
+        nl_index_i1_j1 = None
+        nl_index_i2_j1 = None
+        weights = None
+        tmp_3 = None
+        bincount_bins = None
         if divide_by_masses:
             geom_mean_mass_i1_i2 = np.sqrt(
                 np.take(masses_i, unique_pairs_i1_i2[:, 0]) * np.take(masses_i, unique_pairs_i1_i2[:, 1])
@@ -656,8 +937,7 @@ class EAM(Calculator):
         term_8 = bsr_matrix((tmp_3_summed, unique_pairs_i1_i2[:, 1], index_ptr), shape=(3*nat, 3*nat))
         if symmetry_check:
             print("check term 8", np.linalg.norm(term_8.todense() - term_8.todense().T))
-        D += term_8
-        return D
+        return term_8
 
     @property
     def cutoff(self):
